package lcsb.mapviewer.persist;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

/**
 * Wrapper to always return a reference to the Spring Application Context from
 * within non-Spring enabled beans. Unlike Spring MVC's
 * WebApplicationContextUtils we do not need a reference to the Servlet context
 * for this. All we need is for this bean to be initialized during application
 * startup.
 */
@Transactional
@Component
public class SpringApplicationContext implements ApplicationContextAware {

  /**
   * Default logger for this class.
   */
  private static Logger logger = LogManager.getLogger(SpringApplicationContext.class);

  /**
   * {@link ApplicationContext} used by the Spring framework.
   */
  private static ApplicationContext context;

  /**
   * This is about the same as context.getBean("beanName"), except it has its own
   * static handle to the Spring context, so calling this method statically will
   * give access to the beans by name in the Spring application context. As in the
   * context.getBean("beanName") call, the caller must cast to the appropriate
   * target class. If the bean does not exist, then a Runtime error will be
   * thrown.
   *
   * @param beanName
   *          the name of the bean to get.
   * @return an Object reference to the named bean.
   */
  public static Object getBean(String beanName) {
    beanName = beanName.substring(0, 1).toLowerCase() + beanName.substring(1);
    return context.getBean(beanName);
  }

  /**
   *
   * @return {@link #context}
   */
  public static ApplicationContext getApplicationContext() {
    return context;
  }

  /**
   * This method is called from within the ApplicationContext once it is done
   * starting up, it will stick a reference to itself into this bean.
   *
   * @param context
   *          a reference to the ApplicationContext.
   */
  @Override
  public void setApplicationContext(ApplicationContext context) {
    logger.debug("Setting application context: " + context.getDisplayName());
    SpringApplicationContext.context = context;
  }

}