package lcsb.mapviewer.web;

import static org.junit.Assert.assertEquals;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.Arrays;

import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;
import org.springframework.transaction.annotation.Transactional;

import com.google.gson.JsonParser;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.model.plugin.Plugin;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.dao.plugin.PluginDao;
import lcsb.mapviewer.services.interfaces.IUserService;

@RunWith(SpringJUnit4ClassRunner.class)
@Transactional
@Rollback
public class PluginControllerIntegrationTest extends ControllerIntegrationTest {

  private static final String TEST_USER_PASSWORD = "test_pass";
  private static final String TEST_USER_LOGIN = "test_user";
  Logger logger = LogManager.getLogger();
  @Autowired
  private PluginDao pluginDao;

  @Autowired
  private IUserService userService;

  private User user;

  @Before
  public void setup() {
    user = createUser(TEST_USER_LOGIN, TEST_USER_PASSWORD);
  }

  @Test
  public void createPublicPluginWithoutPrivileges() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("hash", "x"),
        new BasicNameValuePair("name", "x"),
        new BasicNameValuePair("version", "x"),
        new BasicNameValuePair("isPublic", "true"),
        new BasicNameValuePair("url", "x"))));

    RequestBuilder request = post("/plugins/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  @Test
  public void createPublicPluginWithPrivileges() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);
    userService.grantUserPrivilege(user, PrivilegeType.IS_ADMIN);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("hash", "x"),
        new BasicNameValuePair("name", "x"),
        new BasicNameValuePair("version", "x"),
        new BasicNameValuePair("isPublic", "true"),
        new BasicNameValuePair("url", "x"))));

    RequestBuilder request = post("/plugins/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
    assertEquals(1, pluginDao.getAll().size());
  }

  @Test
  public void createPrivatePlugin() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("hash", "x"),
        new BasicNameValuePair("name", "x"),
        new BasicNameValuePair("version", "x"),
        new BasicNameValuePair("isPublic", "false"),
        new BasicNameValuePair("url", "x"))));

    RequestBuilder request = post("/plugins/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
    assertEquals(1, pluginDao.getAll().size());
  }

  @Ignore("API has changed. The user parameter is now implicit, therefor this call is not handled by the servlet.")
  @Test
  public void setAndGetPluginData() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = "value=xxx";

    Plugin plugin = createPlugin();

    RequestBuilder request = post("/plugins/" + plugin.getHash() + "/data/users/" + TEST_USER_LOGIN + "/key/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    request = get("/plugins/" + plugin.getHash() + "/data/users/" + TEST_USER_LOGIN + "/key/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    String result = new JsonParser()
        .parse(response)
        .getAsJsonObject()
        .get("value")
        .getAsString();

    assertEquals("xxx", result);
  }

  @Test
  public void setAndGetGlobalPluginData() throws Exception {
    String body = "value=xxx";
    Plugin plugin = createPlugin();

    RequestBuilder request = post("/plugins/" + plugin.getHash() + "/data/global/key/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);
    request = get("/plugins/" + plugin.getHash() + "/data/global/key/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    String result = new JsonParser()
        .parse(response)
        .getAsJsonObject()
        .get("value")
        .getAsString();

    assertEquals("xxx", result);
  }

  @Ignore("API has changed. The user parameter is now implicit, therefor this call is not handled by the servlet.")
  @Test
  public void forbidAccessingOtherUserPluginData() throws Exception {
    String body = "value=xxx";

    Plugin plugin = createPlugin();

    RequestBuilder request = post(
        "/plugins/" + plugin.getHash() + "/data/users/" + Configuration.ANONYMOUS_LOGIN + "/key/")
            .contentType(MediaType.APPLICATION_FORM_URLENCODED)
            .content(body);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);
    request = get("/plugins/" + plugin.getHash() + "/data/users/" + Configuration.ANONYMOUS_LOGIN + "/key/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isForbidden());
  }

  private Plugin createPlugin() {
    Plugin plugin = new Plugin();
    plugin.setHash("XYZ");
    plugin.setName("a");
    plugin.setVersion("0.0.1");
    pluginDao.add(plugin);
    return plugin;
  }

}
